from typing import Optional
import torch
import torch.nn.functional as F
from skimage.measure import label
import numpy as np
from scipy.ndimage import binary_fill_holes, binary_closing
from GANDLF.utils.generic import get_array_from_image_or_tensor


def torch_morphological(
    input_image, kernel_size: Optional[int] = 1, mode: Optional[str] = "dilation"
) -> torch.Tensor:
    """
    This function performs morphological operations on the input image. Adapted from https://github.com/DIVA-DIA/Generating-Synthetic-Handwritten-Historical-Documents/blob/e6a798dc2b374f338804222747c56cb44869af5b/HTR_ctc/utils/auxilary_functions.py#L10.

    Args:
        input_image (_type_): The input image.
        kernel_size (Optional[int], optional): The size of the window to take a max over.
        mode (Optional[str], optional): The mode of the morphological operation.

    Returns:
        torch.Tensor: The output image after morphological operations.
    """
    assert mode in ["dilation", "erosion", "closing", "opening"], "Invalid mode."
    assert len(input_image.shape) in [
        4,
        5,
    ], "Invalid input shape for morphological operations."
    if len(input_image.shape) == 4:
        max_pool = F.max_pool2d
    elif len(input_image.shape) == 5:
        max_pool = F.max_pool3d

    if mode == "dilation":
        output_image = max_pool(
            input_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )
    elif mode == "erosion":
        output_image = -max_pool(
            -input_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )
    elif mode == "closing":
        output_image = max_pool(
            input_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )
        output_image = -max_pool(
            -output_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )
    elif mode == "opening":
        output_image = -max_pool(
            -input_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )
        output_image = max_pool(
            output_image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2
        )

    return output_image


def fill_holes(
    input_image: torch.Tensor, params: Optional[dict] = None
) -> torch.Tensor:
    """
    This function fills holes in masks.

    Args:
        input_image (torch.Tensor): The input image.
        params (Optional[dict], optional): The parameters dict. Defaults to None.

    Returns:
        torch.Tensor: The output image after morphological operations.
    """
    input_image_array = get_array_from_image_or_tensor(input_image).astype(int)
    input_image_array_closed = binary_closing(input_image_array)
    # Fill the holes in binary objects
    output_array = binary_fill_holes(input_image_array_closed).astype(int)

    return torch.from_numpy(output_array)


def cca(input_image: torch.Tensor, params: Optional[dict] = None) -> torch.Tensor:
    """
    This function performs connected component analysis on the input image.

    Args:
        input_image (torch.Tensor): The input image.
        params (Optional[dict], optional): The parameters dict. Defaults to None.

    Returns:
        torch.Tensor: The output image after morphological operations.
    """
    seg = get_array_from_image_or_tensor(input_image)
    mask = seg != 0

    connectivity = input_image.ndim - 1
    labels_connected = label(mask, connectivity=connectivity)
    labels_connected_sizes = [
        np.sum(labels_connected == i) for i in np.unique(labels_connected)
    ]
    largest_region = 0
    if len(labels_connected_sizes) > 1:
        largest_region = np.argmax(labels_connected_sizes[1:]) + 1
    seg[labels_connected != largest_region] = 0

    return seg



from typing import Union
import torch
import torch.nn.functional as F


def get_modelbase_final_layer(final_convolution_layer: str) -> Union[object, None]:
    """
    This function returns the final convolution layer based on the input string.

    Args:
        final_convolution_layer (str): The string representing the final convolution layer.

    Returns:
        Union[object, None]: The final convolution layer.
    """
    none_list = [
        "none",
        None,
        "None",
        "regression",
        "classification_but_not_softmax",
        "logits",
        "classification_without_softmax",
    ]

    if final_convolution_layer in ["sigmoid", "sig"]:
        final_convolution_layer = torch.sigmoid

    elif final_convolution_layer in ["softmax", "soft"]:
        final_convolution_layer = F.softmax

    elif final_convolution_layer in none_list:
        final_convolution_layer = None

    return final_convolution_layer



import torch.nn as nn


class IncDropout(nn.Module):
    def __init__(
        self,
        input_channels: int,
        output_channels: int,
        Conv: nn.Module = nn.Conv2d,
        Dropout: nn.Module = nn.Dropout2d,
        InstanceNorm: nn.Module = nn.InstanceNorm2d,
        dropout_p: float = 0.3,
        leakiness: float = 1e-2,
        conv_bias: bool = True,
        inst_norm_affine: bool = True,
        res: bool = False,
        lrelu_inplace: bool = True,
    ):
        """
        Incremental dropout module.

        Args:
            input_channels (int): Number of input channels.
            output_channels (int): Number of output channels.
            Conv (nn.Module, optional): The convolutional layer type. Defaults to nn.Conv2d.
            Dropout (nn.Module, optional): The dropout layer type. Defaults to nn.Dropout2d.
            InstanceNorm (nn.Module, optional): The instance normalization layer type. Defaults to nn.InstanceNorm2d.
            dropout_p (float, optional): The dropout probability. Defaults to 0.3.
            leakiness (float, optional): The leakiness of the leaky ReLU. Defaults to 1e-2.
            conv_bias (bool, optional): The bias in the convolutional layer. Defaults to True.
            inst_norm_affine (bool, optional): Whether to use the affine transformation in the instance normalization layer. Defaults to True.
            res (bool, optional): Whether to use residual connections. Defaults to False.
            lrelu_inplace (bool, optional): Whether to use the inplace version of the leaky ReLU. Defaults to True.
        """
        nn.Module.__init__(self)

        self.dropout = Dropout(dropout_p)
        self.conv = Conv(
            input_channels,
            output_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=conv_bias,
        )

    def forward(self, x):
        """
        Forward pass of the incremental dropout module.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            x (torch.Tensor): The output tensor.
        """
        x = self.dropout(x)
        x = self.conv(x)
        return x
